@@ -1,5 +1,5 @@
 /*
-Copyright 2016 Google Inc. All Rights Reserved.
+Copyright 2016 Google LLC
 
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
@@ -16,37 +16,39 @@ limitations under the License.
 package bigtable
 
 import (
-	"reflect"
+	"context"
 	"strings"
 	"testing"
 	"time"
 
 	"cloud.google.com/go/bigtable/bttest"
+	"cloud.google.com/go/internal/testutil"
 	"github.com/golang/protobuf/ptypes/wrappers"
-	"golang.org/x/net/context"
+	"github.com/google/go-cmp/cmp"
 	"google.golang.org/api/option"
 	btpb "google.golang.org/genproto/googleapis/bigtable/v2"
 	rpcpb "google.golang.org/genproto/googleapis/rpc/status"
 	"google.golang.org/grpc"
 	"google.golang.org/grpc/codes"
+	"google.golang.org/grpc/status"
 )
 
 func setupFakeServer(opt ...grpc.ServerOption) (tbl *Table, cleanup func(), err error) {
-	srv, err := bttest.NewServer("127.0.0.1:0", opt...)
+	srv, err := bttest.NewServer("localhost:0", opt...)
 	if err != nil {
 		return nil, nil, err
 	}
-	conn, err := grpc.Dial(srv.Addr, grpc.WithInsecure())
+	conn, err := grpc.Dial(srv.Addr, grpc.WithInsecure(), grpc.WithBlock())
 	if err != nil {
 		return nil, nil, err
 	}
 
-	client, err := NewClient(context.Background(), "client", "instance", option.WithGRPCConn(conn))
+	client, err := NewClient(context.Background(), "client", "instance", option.WithGRPCConn(conn), option.WithGRPCDialOption(grpc.WithBlock()))
 	if err != nil {
 		return nil, nil, err
 	}
 
-	adminClient, err := NewAdminClient(context.Background(), "client", "instance", option.WithGRPCConn(conn))
+	adminClient, err := NewAdminClient(context.Background(), "client", "instance", option.WithGRPCConn(conn), option.WithGRPCDialOption(grpc.WithBlock()))
 	if err != nil {
 		return nil, nil, err
 	}
@@ -75,18 +77,18 @@ func TestRetryApply(t *testing.T) {
 	errInjector := func(ctx context.Context, req interface{}, info *grpc.UnaryServerInfo, handler grpc.UnaryHandler) (interface{}, error) {
 		if strings.HasSuffix(info.FullMethod, "MutateRow") && errCount < 3 {
 			errCount++
-			return nil, grpc.Errorf(code, "")
+			return nil, status.Errorf(code, "")
 		}
 		return handler(ctx, req)
 	}
 	tbl, cleanup, err := setupFakeServer(grpc.UnaryInterceptor(errInjector))
-	defer cleanup()
 	if err != nil {
 		t.Fatalf("fake server setup: %v", err)
 	}
+	defer cleanup()
 
 	mut := NewMutation()
-	mut.Set("cf", "col", 1, []byte("val"))
+	mut.Set("cf", "col", 1000, []byte("val"))
 	if err := tbl.Apply(ctx, "row1", mut); err != nil {
 		t.Errorf("applying single mutation with retries: %v", err)
 	}
@@ -108,8 +110,8 @@ func TestRetryApply(t *testing.T) {
 	mutTrue := NewMutation()
 	mutTrue.DeleteRow()
 	mutFalse := NewMutation()
-	mutFalse.Set("cf", "col", 1, []byte("val"))
-	condMut := NewCondMutation(ValueFilter("."), mutTrue, mutFalse)
+	mutFalse.Set("cf", "col", 1000, []byte("val"))
+	condMut := NewCondMutation(ValueFilter(".*"), mutTrue, mutFalse)
 
 	errCount = 0
 	code = codes.Unavailable // Will be retried
@@ -131,15 +133,21 @@ func TestRetryApply(t *testing.T) {
 	}
 }
 
-func TestRetryApplyBulk(t *testing.T) {
+// Test overall request failure and retries.
+func TestRetryApplyBulk_OverallRequestFailure(t *testing.T) {
 	ctx := context.Background()
 
 	// Intercept requests and delegate to an interceptor defined by the test case
 	errCount := 0
-	var f func(grpc.ServerStream) error
 	errInjector := func(srv interface{}, ss grpc.ServerStream, info *grpc.StreamServerInfo, handler grpc.StreamHandler) error {
 		if strings.HasSuffix(info.FullMethod, "MutateRows") {
-			return f(ss)
+			return func() error {
+				if errCount < 3 {
+					errCount++
+					return status.Errorf(codes.Aborted, "")
+				}
+				return nil
+			}()
 		}
 		return handler(ctx, ss)
 	}
@@ -151,22 +159,64 @@ func TestRetryApplyBulk(t *testing.T) {
 	}
 
 	errCount = 0
-	// Test overall request failure and retries
-	f = func(ss grpc.ServerStream) error {
-		if errCount < 3 {
-			errCount++
-			return grpc.Errorf(codes.Aborted, "")
-		}
-		return nil
-	}
+
 	mut := NewMutation()
 	mut.Set("cf", "col", 1, []byte{})
 	errors, err := tbl.ApplyBulk(ctx, []string{"row2"}, []*Mutation{mut})
 	if errors != nil || err != nil {
 		t.Errorf("bulk with request failure: got: %v, %v, want: nil", errors, err)
 	}
+}
+
+func TestRetryApplyBulk_FailuresAndRetriesInOneRequest(t *testing.T) {
+	ctx := context.Background()
+
+	// Intercept requests and delegate to an interceptor defined by the test case
+	errCount := 0
+	errInjector := func(srv interface{}, ss grpc.ServerStream, info *grpc.StreamServerInfo, handler grpc.StreamHandler) error {
+		if strings.HasSuffix(info.FullMethod, "MutateRows") {
+			return func(ss grpc.ServerStream) error {
+				var err error
+				req := new(btpb.MutateRowsRequest)
+				must(ss.RecvMsg(req))
+				switch errCount {
+				case 0:
+					// Retryable request failure
+					err = status.Errorf(codes.Unavailable, "")
+				case 1:
+					// Two mutations fail
+					must(writeMutateRowsResponse(ss, codes.Unavailable, codes.OK, codes.Aborted))
+					err = nil
+				case 2:
+					// Two failures were retried. One will succeed.
+					if want, got := 2, len(req.Entries); want != got {
+						t.Fatalf("2 bulk retries, got: %d, want %d", got, want)
+					}
+					must(writeMutateRowsResponse(ss, codes.OK, codes.Aborted))
+					err = nil
+				case 3:
+					// One failure was retried and will succeed.
+					if want, got := 1, len(req.Entries); want != got {
+						t.Fatalf("1 bulk retry, got: %d, want %d", got, want)
+					}
+					must(writeMutateRowsResponse(ss, codes.OK))
+					err = nil
+				}
+				errCount++
+				return err
+			}(ss)
+		}
+		return handler(ctx, ss)
+	}
+
+	tbl, cleanup, err := setupFakeServer(grpc.StreamInterceptor(errInjector))
+	defer cleanup()
+	if err != nil {
+		t.Fatalf("fake server setup: %v", err)
+	}
+
+	errCount = 0
 
-	// Test failures and retries in one request
 	errCount = 0
 	m1 := NewMutation()
 	m1.Set("cf", "col", 1, []byte{})
@@ -174,83 +224,107 @@ func TestRetryApplyBulk(t *testing.T) {
 	m2.Set("cf", "col2", 1, []byte{})
 	m3 := NewMutation()
 	m3.Set("cf", "col3", 1, []byte{})
-	f = func(ss grpc.ServerStream) error {
-		var err error
-		req := new(btpb.MutateRowsRequest)
-		ss.RecvMsg(req)
-		switch errCount {
-		case 0:
-			// Retryable request failure
-			err = grpc.Errorf(codes.Unavailable, "")
-		case 1:
-			// Two mutations fail
-			writeMutateRowsResponse(ss, codes.Unavailable, codes.OK, codes.Aborted)
-			err = nil
-		case 2:
-			// Two failures were retried. One will succeed.
-			if want, got := 2, len(req.Entries); want != got {
-				t.Errorf("2 bulk retries, got: %d, want %d", got, want)
-			}
-			writeMutateRowsResponse(ss, codes.OK, codes.Aborted)
-			err = nil
-		case 3:
-			// One failure was retried and will succeed.
-			if want, got := 1, len(req.Entries); want != got {
-				t.Errorf("1 bulk retry, got: %d, want %d", got, want)
-			}
-			writeMutateRowsResponse(ss, codes.OK)
-			err = nil
-		}
-		errCount++
-		return err
-	}
-	errors, err = tbl.ApplyBulk(ctx, []string{"row1", "row2", "row3"}, []*Mutation{m1, m2, m3})
+	errors, err := tbl.ApplyBulk(ctx, []string{"row1", "row2", "row3"}, []*Mutation{m1, m2, m3})
 	if errors != nil || err != nil {
 		t.Errorf("bulk with retries: got: %v, %v, want: nil", errors, err)
 	}
+}
+
+func TestRetryApplyBulk_UnretryableErrors(t *testing.T) {
+	ctx := context.Background()
+
+	// Intercept requests and delegate to an interceptor defined by the test case
+	errCount := 0
+	errInjector := func(srv interface{}, ss grpc.ServerStream, info *grpc.StreamServerInfo, handler grpc.StreamHandler) error {
+		if strings.HasSuffix(info.FullMethod, "MutateRows") {
+			return func(ss grpc.ServerStream) error {
+				var err error
+				req := new(btpb.MutateRowsRequest)
+				must(ss.RecvMsg(req))
+				switch errCount {
+				case 0:
+					// Give non-idempotent mutation a retryable error code.
+					// Nothing should be retried.
+					must(writeMutateRowsResponse(ss, codes.FailedPrecondition, codes.Aborted))
+					err = nil
+				case 1:
+					t.Fatalf("unretryable errors: got one retry, want no retries")
+				}
+				errCount++
+				return err
+			}(ss)
+		}
+		return handler(ctx, ss)
+	}
+
+	tbl, cleanup, err := setupFakeServer(grpc.StreamInterceptor(errInjector))
+	defer cleanup()
+	if err != nil {
+		t.Fatalf("fake server setup: %v", err)
+	}
+
+	errCount = 0
+
+	m1 := NewMutation()
+	m1.Set("cf", "col", 1, []byte{})
+	m2 := NewMutation()
+	m2.Set("cf", "col2", 1, []byte{})
+	m3 := NewMutation()
+	m3.Set("cf", "col3", 1, []byte{})
 
-	// Test unretryable errors
 	niMut := NewMutation()
 	niMut.Set("cf", "col", ServerTime, []byte{}) // Non-idempotent
 	errCount = 0
-	f = func(ss grpc.ServerStream) error {
-		var err error
-		req := new(btpb.MutateRowsRequest)
-		ss.RecvMsg(req)
-		switch errCount {
-		case 0:
-			// Give non-idempotent mutation a retryable error code.
-			// Nothing should be retried.
-			writeMutateRowsResponse(ss, codes.FailedPrecondition, codes.Aborted)
-			err = nil
-		case 1:
-			t.Errorf("unretryable errors: got one retry, want no retries")
-		}
-		errCount++
-		return err
-	}
-	errors, err = tbl.ApplyBulk(ctx, []string{"row1", "row2"}, []*Mutation{m1, niMut})
+	errors, err := tbl.ApplyBulk(ctx, []string{"row1", "row2"}, []*Mutation{m1, niMut})
 	if err != nil {
-		t.Errorf("unretryable errors: request failed %v")
+		t.Fatalf("unretryable errors: request failed %v", err)
 	}
 	want := []error{
-		grpc.Errorf(codes.FailedPrecondition, ""),
-		grpc.Errorf(codes.Aborted, ""),
+		status.Errorf(codes.FailedPrecondition, ""),
+		status.Errorf(codes.Aborted, ""),
 	}
-	if !reflect.DeepEqual(want, errors) {
+	if !testutil.Equal(want, errors,
+		cmp.Comparer(func(x, y error) bool {
+			return x == y || (x != nil && y != nil && x.Error() == y.Error())
+		}),
+	) {
 		t.Errorf("unretryable errors: got: %v, want: %v", errors, want)
 	}
+}
 
-	// Test individual errors and a deadline exceeded
-	f = func(ss grpc.ServerStream) error {
-		writeMutateRowsResponse(ss, codes.FailedPrecondition, codes.OK, codes.Aborted)
-		return nil
+func TestRetryApplyBulk_IndividualErrorsAndDeadlineExceeded(t *testing.T) {
+	ctx := context.Background()
+
+	// Intercept requests and delegate to an interceptor defined by the test case
+	errInjector := func(srv interface{}, ss grpc.ServerStream, info *grpc.StreamServerInfo, handler grpc.StreamHandler) error {
+		if strings.HasSuffix(info.FullMethod, "MutateRows") {
+			return func(ss grpc.ServerStream) error {
+				return writeMutateRowsResponse(ss, codes.FailedPrecondition, codes.OK, codes.Aborted)
+			}(ss)
+		}
+		return handler(ctx, ss)
 	}
-	ctx, _ = context.WithTimeout(ctx, 100*time.Millisecond)
-	errors, err = tbl.ApplyBulk(ctx, []string{"row1", "row2", "row3"}, []*Mutation{m1, m2, m3})
+
+	tbl, cleanup, err := setupFakeServer(grpc.StreamInterceptor(errInjector))
+	defer cleanup()
+	if err != nil {
+		t.Fatalf("fake server setup: %v", err)
+	}
+
+	m1 := NewMutation()
+	m1.Set("cf", "col", 1, []byte{})
+	m2 := NewMutation()
+	m2.Set("cf", "col2", 1, []byte{})
+	m3 := NewMutation()
+	m3.Set("cf", "col3", 1, []byte{})
+
+	// This should cause a deadline exceeded error.
+	ctx, cancel := context.WithTimeout(ctx, -10*time.Millisecond)
+	defer cancel()
+	errors, err := tbl.ApplyBulk(ctx, []string{"row1", "row2", "row3"}, []*Mutation{m1, m2, m3})
 	wantErr := context.DeadlineExceeded
 	if wantErr != err {
-		t.Errorf("deadline exceeded error: got: %v, want: %v", err, wantErr)
+		t.Fatalf("deadline exceeded error: got: %v, want: %v", err, wantErr)
 	}
 	if errors != nil {
 		t.Errorf("deadline exceeded errors: got: %v, want: nil", err)
@@ -273,7 +347,7 @@ func TestRetainRowsAfter(t *testing.T) {
 	prevRowKey := "m"
 	want := NewRange("m\x00", "z")
 	got := prevRowRange.retainRowsAfter(prevRowKey)
-	if !reflect.DeepEqual(want, got) {
+	if !testutil.Equal(want, got, cmp.AllowUnexported(RowRange{})) {
 		t.Errorf("range retry: got %v, want %v", got, want)
 	}
 
@@ -281,7 +355,7 @@ func TestRetainRowsAfter(t *testing.T) {
 	prevRowKey = "f"
 	wantRowRangeList := RowRangeList{NewRange("f\x00", "g"), NewRange("h", "l")}
 	got = prevRowRangeList.retainRowsAfter(prevRowKey)
-	if !reflect.DeepEqual(wantRowRangeList, got) {
+	if !testutil.Equal(wantRowRangeList, got, cmp.AllowUnexported(RowRange{})) {
 		t.Errorf("range list retry: got %v, want %v", got, wantRowRangeList)
 	}
 
@@ -289,7 +363,7 @@ func TestRetainRowsAfter(t *testing.T) {
 	prevRowKey = "b"
 	wantList := RowList{"c", "d", "e", "f"}
 	got = prevRowList.retainRowsAfter(prevRowKey)
-	if !reflect.DeepEqual(wantList, got) {
+	if !testutil.Equal(wantList, got) {
 		t.Errorf("list retry: got %v, want %v", got, wantList)
 	}
 }
@@ -318,27 +392,27 @@ func TestRetryReadRows(t *testing.T) {
 	f = func(ss grpc.ServerStream) error {
 		var err error
 		req := new(btpb.ReadRowsRequest)
-		ss.RecvMsg(req)
+		must(ss.RecvMsg(req))
 		switch errCount {
 		case 0:
 			// Retryable request failure
-			err = grpc.Errorf(codes.Unavailable, "")
+			err = status.Errorf(codes.Unavailable, "")
 		case 1:
 			// Write two rows then error
 			if want, got := "a", string(req.Rows.RowRanges[0].GetStartKeyClosed()); want != got {
 				t.Errorf("first retry, no data received yet: got %q, want %q", got, want)
 			}
-			writeReadRowsResponse(ss, "a", "b")
-			err = grpc.Errorf(codes.Unavailable, "")
+			must(writeReadRowsResponse(ss, "a", "b"))
+			err = status.Errorf(codes.Unavailable, "")
 		case 2:
 			// Retryable request failure
 			if want, got := "b\x00", string(req.Rows.RowRanges[0].GetStartKeyClosed()); want != got {
 				t.Errorf("2 range retries: got %q, want %q", got, want)
 			}
-			err = grpc.Errorf(codes.Unavailable, "")
+			err = status.Errorf(codes.Unavailable, "")
 		case 3:
 			// Write two more rows
-			writeReadRowsResponse(ss, "c", "d")
+			must(writeReadRowsResponse(ss, "c", "d"))
 			err = nil
 		}
 		errCount++
@@ -346,12 +420,12 @@ func TestRetryReadRows(t *testing.T) {
 	}
 
 	var got []string
-	tbl.ReadRows(ctx, NewRange("a", "z"), func(r Row) bool {
+	must(tbl.ReadRows(ctx, NewRange("a", "z"), func(r Row) bool {
 		got = append(got, r.Key())
 		return true
-	})
+	}))
 	want := []string{"a", "b", "c", "d"}
-	if !reflect.DeepEqual(got, want) {
+	if !testutil.Equal(got, want) {
 		t.Errorf("retry range integration: got %v, want %v", got, want)
 	}
 }
@@ -368,3 +442,9 @@ func writeReadRowsResponse(ss grpc.ServerStream, rowKeys ...string) error {
 	}
 	return ss.SendMsg(&btpb.ReadRowsResponse{Chunks: chunks})
 }
+
+func must(err error) {
+	if err != nil {
+		panic(err)
+	}
+}
