@@ -1,4 +1,4 @@
-// Copyright 2016 Google Inc. All Rights Reserved.
+// Copyright 2016 Google LLC
 //
 // Licensed under the Apache License, Version 2.0 (the "License");
 // you may not use this file except in compliance with the License.
@@ -12,11 +12,12 @@
 // See the License for the specific language governing permissions and
 // limitations under the License.
 
-// +build linux
+// +build linux,go1.7
 
 package main
 
 import (
+	"context"
 	"encoding/json"
 	"flag"
 	"fmt"
@@ -29,11 +30,10 @@ import (
 
 	"cloud.google.com/go/cmd/go-cloud-debug-agent/internal/breakpoints"
 	debuglet "cloud.google.com/go/cmd/go-cloud-debug-agent/internal/controller"
+	"cloud.google.com/go/cmd/go-cloud-debug-agent/internal/debug"
+	"cloud.google.com/go/cmd/go-cloud-debug-agent/internal/debug/local"
 	"cloud.google.com/go/cmd/go-cloud-debug-agent/internal/valuecollector"
 	"cloud.google.com/go/compute/metadata"
-	"golang.org/x/debug"
-	"golang.org/x/debug/local"
-	"golang.org/x/net/context"
 	"golang.org/x/oauth2"
 	"golang.org/x/oauth2/google"
 	cd "google.golang.org/api/clouddebugger/v2"
