@@ -1,4 +1,4 @@
-// Copyright 2014 Google Inc. All Rights Reserved.
+// Copyright 2014 Google LLC
 //
 // Licensed under the Apache License, Version 2.0 (the "License");
 // you may not use this file except in compliance with the License.
@@ -14,36 +14,46 @@
 
 // Package container contains a deprecated Google Container Engine client.
 //
-// Deprecated: Use google.golang.org/api/container instead.
+// Deprecated: Use cloud.google.com/go/container/apiv1 instead.
 package container // import "cloud.google.com/go/container"
 
 import (
+	"context"
 	"errors"
 	"fmt"
 	"time"
 
-	"golang.org/x/net/context"
 	raw "google.golang.org/api/container/v1"
 	"google.golang.org/api/option"
-	"google.golang.org/api/transport"
+	htransport "google.golang.org/api/transport/http"
 )
 
+// Type is the operation type.
 type Type string
 
 const (
+	// TypeCreate describes that the operation is "create cluster".
 	TypeCreate = Type("createCluster")
+	// TypeDelete describes that the operation is "delete cluster".
 	TypeDelete = Type("deleteCluster")
 )
 
+// Status is the current status of the operation or resource.
 type Status string
 
 const (
-	StatusDone         = Status("done")
-	StatusPending      = Status("pending")
-	StatusRunning      = Status("running")
-	StatusError        = Status("error")
+	// StatusDone is a status indicating that the resource or operation is in done state.
+	StatusDone = Status("done")
+	// StatusPending is a status indicating that the resource or operation is in pending state.
+	StatusPending = Status("pending")
+	// StatusRunning is a status indicating that the resource or operation is in running state.
+	StatusRunning = Status("running")
+	// StatusError is a status indicating that the resource or operation is in error state.
+	StatusError = Status("error")
+	// StatusProvisioning is a status indicating that the resource or operation is in provisioning state.
 	StatusProvisioning = Status("provisioning")
-	StatusStopping     = Status("stopping")
+	// StatusStopping is a status indicating that the resource or operation is in stopping state.
+	StatusStopping = Status("stopping")
 )
 
 const prodAddr = "https://container.googleapis.com/"
@@ -64,7 +74,7 @@ func NewClient(ctx context.Context, projectID string, opts ...option.ClientOptio
 		option.WithUserAgent(userAgent),
 	}
 	o = append(o, opts...)
-	httpClient, endpoint, err := transport.NewHTTPClient(ctx, o...)
+	httpClient, endpoint, err := htransport.NewClient(ctx, o...)
 	if err != nil {
 		return nil, fmt.Errorf("dialing: %v", err)
 	}
@@ -229,12 +239,6 @@ func (c *Client) Cluster(ctx context.Context, zone, name string) (*Resource, err
 	return resourceFromRaw(resp), nil
 }
 
-// CreateCluster creates a new cluster with the provided metadata
-// in the specified zone.
-func (c *Client) CreateCluster(ctx context.Context, zone string, resource *Resource) (*Resource, error) {
-	panic("not implemented")
-}
-
 // DeleteCluster deletes a cluster.
 func (c *Client) DeleteCluster(ctx context.Context, zone, name string) error {
 	_, err := c.svc.Projects.Zones.Clusters.Delete(c.projectID, zone, name).Do()
