@@ -1,4 +1,4 @@
-// Copyright 2014 Google Inc. All Rights Reserved.
+// Copyright 2014 Google LLC
 //
 // Licensed under the Apache License, Version 2.0 (the "License");
 // you may not use this file except in compliance with the License.
@@ -15,8 +15,13 @@
 package datastore
 
 import (
+	"context"
+	"encoding/json"
 	"errors"
+	"flag"
 	"fmt"
+	"log"
+	"os"
 	"reflect"
 	"sort"
 	"strings"
@@ -25,46 +30,157 @@ import (
 	"time"
 
 	"cloud.google.com/go/internal/testutil"
-	"golang.org/x/net/context"
+	"cloud.google.com/go/internal/uid"
+	"cloud.google.com/go/rpcreplay"
 	"google.golang.org/api/iterator"
 	"google.golang.org/api/option"
+	"google.golang.org/grpc"
+	"google.golang.org/grpc/codes"
+	"google.golang.org/grpc/status"
 )
 
 // TODO(djd): Make test entity clean up more robust: some test entities may
 // be left behind if tests are aborted, the transport fails, etc.
 
+var timeNow = time.Now()
+
 // suffix is a timestamp-based suffix which is appended to key names,
 // particularly for the root keys of entity groups. This reduces flakiness
 // when the tests are run in parallel.
-var suffix = fmt.Sprintf("-t%d", time.Now().UnixNano())
+var suffix string
 
-func newClient(ctx context.Context, t *testing.T) *Client {
+const replayFilename = "datastore.replay"
+
+type replayInfo struct {
+	ProjectID string
+	Time      time.Time
+}
+
+var (
+	record = flag.Bool("record", false, "record RPCs")
+
+	newTestClient = func(ctx context.Context, t *testing.T) *Client {
+		return newClient(ctx, t, nil)
+	}
+)
+
+func TestMain(m *testing.M) {
+	os.Exit(testMain(m))
+}
+
+func testMain(m *testing.M) int {
+	flag.Parse()
+	if testing.Short() {
+		if *record {
+			log.Fatal("cannot combine -short and -record")
+		}
+		if testutil.CanReplay(replayFilename) {
+			initReplay()
+		}
+	} else if *record {
+		if testutil.ProjID() == "" {
+			log.Fatal("must record with a project ID")
+		}
+		b, err := json.Marshal(replayInfo{
+			ProjectID: testutil.ProjID(),
+			Time:      timeNow,
+		})
+		if err != nil {
+			log.Fatal(err)
+		}
+		rec, err := rpcreplay.NewRecorder(replayFilename, b)
+		if err != nil {
+			log.Fatal(err)
+		}
+		defer func() {
+			if err := rec.Close(); err != nil {
+				log.Fatalf("closing recorder: %v", err)
+			}
+		}()
+		newTestClient = func(ctx context.Context, t *testing.T) *Client {
+			return newClient(ctx, t, rec.DialOptions())
+		}
+		log.Printf("recording to %s", replayFilename)
+	}
+	suffix = fmt.Sprintf("-t%d", timeNow.UnixNano())
+	return m.Run()
+}
+
+func initReplay() {
+	rep, err := rpcreplay.NewReplayer(replayFilename)
+	if err != nil {
+		log.Fatal(err)
+	}
+	defer rep.Close()
+
+	var ri replayInfo
+	if err := json.Unmarshal(rep.Initial(), &ri); err != nil {
+		log.Fatalf("unmarshaling initial replay info: %v", err)
+	}
+	timeNow = ri.Time.In(time.Local)
+
+	conn, err := rep.Connection()
+	if err != nil {
+		log.Fatal(err)
+	}
+
+	newTestClient = func(ctx context.Context, t *testing.T) *Client {
+		grpcHeadersEnforcer := &testutil.HeadersEnforcer{
+			OnFailure: t.Fatalf,
+			Checkers: []*testutil.HeaderChecker{
+				testutil.XGoogClientHeaderChecker,
+			},
+		}
+
+		opts := append(grpcHeadersEnforcer.CallOptions(), option.WithGRPCConn(conn))
+		client, err := NewClient(ctx, ri.ProjectID, opts...)
+		if err != nil {
+			t.Fatalf("NewClient: %v", err)
+		}
+		return client
+	}
+	log.Printf("replaying from %s", replayFilename)
+}
+
+func newClient(ctx context.Context, t *testing.T, dialOpts []grpc.DialOption) *Client {
+	if testing.Short() {
+		t.Skip("Integration tests skipped in short mode")
+	}
 	ts := testutil.TokenSource(ctx, ScopeDatastore)
 	if ts == nil {
 		t.Skip("Integration tests skipped. See CONTRIBUTING.md for details")
 	}
-	client, err := NewClient(ctx, testutil.ProjID(), option.WithTokenSource(ts))
+
+	grpcHeadersEnforcer := &testutil.HeadersEnforcer{
+		OnFailure: t.Fatalf,
+		Checkers: []*testutil.HeaderChecker{
+			testutil.XGoogClientHeaderChecker,
+		},
+	}
+	opts := append(grpcHeadersEnforcer.CallOptions(), option.WithTokenSource(ts))
+	for _, opt := range dialOpts {
+		opts = append(opts, option.WithGRPCDialOption(opt))
+	}
+	client, err := NewClient(ctx, testutil.ProjID(), opts...)
 	if err != nil {
 		t.Fatalf("NewClient: %v", err)
 	}
 	return client
 }
 
-func TestBasics(t *testing.T) {
-	if testing.Short() {
-		t.Skip("Integration tests skipped in short mode")
-	}
+func TestIntegration_Basics(t *testing.T) {
 	ctx, _ := context.WithTimeout(context.Background(), time.Second*20)
-	client := newClient(ctx, t)
+	client := newTestClient(ctx, t)
 	defer client.Close()
 
 	type X struct {
 		I int
 		S string
 		T time.Time
+		U interface{}
 	}
 
-	x0 := X{66, "99", time.Now().Truncate(time.Millisecond)}
+	x0 := X{66, "99", timeNow.Truncate(time.Millisecond), "X"}
 	k, err := client.Put(ctx, IncompleteKey("BasicsX", nil), &x0)
 	if err != nil {
 		t.Fatalf("client.Put: %v", err)
@@ -78,18 +194,14 @@ func TestBasics(t *testing.T) {
 	if err != nil {
 		t.Errorf("client.Delete: %v", err)
 	}
-	if !reflect.DeepEqual(x0, x1) {
+	if !testutil.Equal(x0, x1) {
 		t.Errorf("compare: x0=%v, x1=%v", x0, x1)
 	}
 }
 
-func TestTopLevelKeyLoaded(t *testing.T) {
-	if testing.Short() {
-		t.Skip("Integration tests skipped in short mode")
-	}
-
+func TestIntegration_TopLevelKeyLoaded(t *testing.T) {
 	ctx, _ := context.WithTimeout(context.Background(), time.Second*20)
-	client := newClient(ctx, t)
+	client := newTestClient(ctx, t)
 	defer client.Close()
 
 	completeKey := NameKey("EntityWithKey", "myent", nil)
@@ -117,18 +229,15 @@ func TestTopLevelKeyLoaded(t *testing.T) {
 	}
 
 	// The two keys should be absolutely identical.
-	if !reflect.DeepEqual(e.K, k) {
+	if !testutil.Equal(e.K, k) {
 		t.Fatalf("e.K not equal to k; got %#v, want %#v", e.K, k)
 	}
 
 }
 
-func TestListValues(t *testing.T) {
-	if testing.Short() {
-		t.Skip("Integration tests skipped in short mode")
-	}
+func TestIntegration_ListValues(t *testing.T) {
 	ctx := context.Background()
-	client := newClient(ctx, t)
+	client := newTestClient(ctx, t)
 	defer client.Close()
 
 	p0 := PropertyList{
@@ -142,7 +251,7 @@ func TestListValues(t *testing.T) {
 	if err := client.Get(ctx, k, &p1); err != nil {
 		t.Errorf("client.Get: %v", err)
 	}
-	if !reflect.DeepEqual(p0, p1) {
+	if !testutil.Equal(p0, p1) {
 		t.Errorf("compare:\np0=%v\np1=%#v", p0, p1)
 	}
 	if err = client.Delete(ctx, k); err != nil {
@@ -150,12 +259,9 @@ func TestListValues(t *testing.T) {
 	}
 }
 
-func TestGetMulti(t *testing.T) {
-	if testing.Short() {
-		t.Skip("Integration tests skipped in short mode")
-	}
+func TestIntegration_GetMulti(t *testing.T) {
 	ctx := context.Background()
-	client := newClient(ctx, t)
+	client := newTestClient(ctx, t)
 	defer client.Close()
 
 	type X struct {
@@ -170,6 +276,7 @@ func TestGetMulti(t *testing.T) {
 		{key: NameKey("X", "item1", p), put: true},
 		{key: NameKey("X", "item2", p), put: false},
 		{key: NameKey("X", "item3", p), put: false},
+		{key: NameKey("X", "item3", p), put: false},
 		{key: NameKey("X", "item4", p), put: true},
 	}
 
@@ -223,12 +330,9 @@ func (z Z) String() string {
 	return fmt.Sprintf("Z{ %s }", strings.Join(lens, ","))
 }
 
-func TestUnindexableValues(t *testing.T) {
-	if testing.Short() {
-		t.Skip("Integration tests skipped in short mode")
-	}
+func TestIntegration_UnindexableValues(t *testing.T) {
 	ctx := context.Background()
-	client := newClient(ctx, t)
+	client := newTestClient(ctx, t)
 	defer client.Close()
 
 	x1500 := strings.Repeat("x", 1500)
@@ -254,12 +358,9 @@ func TestUnindexableValues(t *testing.T) {
 	}
 }
 
-func TestNilKey(t *testing.T) {
-	if testing.Short() {
-		t.Skip("Integration tests skipped in short mode")
-	}
+func TestIntegration_NilKey(t *testing.T) {
 	ctx := context.Background()
-	client := newClient(ctx, t)
+	client := newTestClient(ctx, t)
 	defer client.Close()
 
 	testCases := []struct {
@@ -289,7 +390,7 @@ type SQTestCase struct {
 	wantSum   int
 }
 
-func testSmallQueries(t *testing.T, ctx context.Context, client *Client, parent *Key, children []*SQChild,
+func testSmallQueries(ctx context.Context, t *testing.T, client *Client, parent *Key, children []*SQChild,
 	testCases []SQTestCase, extraTests ...func()) {
 	keys := make([]*Key, len(children))
 	for i := range keys {
@@ -339,16 +440,13 @@ func testSmallQueries(t *testing.T, ctx context.Context, client *Client, parent
 	}
 }
 
-func TestFilters(t *testing.T) {
-	if testing.Short() {
-		t.Skip("Integration tests skipped in short mode")
-	}
+func TestIntegration_Filters(t *testing.T) {
 	ctx := context.Background()
-	client := newClient(ctx, t)
+	client := newTestClient(ctx, t)
 	defer client.Close()
 
-	parent := NameKey("SQParent", "TestFilters"+suffix, nil)
-	now := time.Now().Truncate(time.Millisecond).Unix()
+	parent := NameKey("SQParent", "TestIntegration_Filters"+suffix, nil)
+	now := timeNow.Truncate(time.Millisecond).Unix()
 	children := []*SQChild{
 		{I: 0, T: now, U: now},
 		{I: 1, T: now, U: now},
@@ -360,7 +458,7 @@ func TestFilters(t *testing.T) {
 		{I: 7, T: now, U: now},
 	}
 	baseQuery := NewQuery("SQChild").Ancestor(parent).Filter("T=", now)
-	testSmallQueries(t, ctx, client, parent, children, []SQTestCase{
+	testSmallQueries(ctx, t, client, parent, children, []SQTestCase{
 		{
 			"I>1",
 			baseQuery.Filter("I>", 1),
@@ -401,7 +499,7 @@ func TestFilters(t *testing.T) {
 		if err != nil {
 			t.Errorf("client.GetAll: %v", err)
 		}
-		if !reflect.DeepEqual(got, want) {
+		if !testutil.Equal(got, want) {
 			t.Errorf("compare: got=%v, want=%v", got, want)
 		}
 	}, func() {
@@ -420,22 +518,21 @@ func TestFilters(t *testing.T) {
 		if err != nil {
 			t.Errorf("client.GetAll: %v", err)
 		}
-		if !reflect.DeepEqual(got, want) {
+		if !testutil.Equal(got, want) {
 			t.Errorf("compare: got=%v, want=%v", got, want)
 		}
 	})
 }
 
-func TestLargeQuery(t *testing.T) {
-	if testing.Short() {
-		t.Skip("Integration tests skipped in short mode")
-	}
+type ckey struct{}
+
+func TestIntegration_LargeQuery(t *testing.T) {
 	ctx := context.Background()
-	client := newClient(ctx, t)
+	client := newTestClient(ctx, t)
 	defer client.Close()
 
-	parent := NameKey("LQParent", "TestFilters"+suffix, nil)
-	now := time.Now().Truncate(time.Millisecond).Unix()
+	parent := NameKey("LQParent", "TestIntegration_LargeQuery"+suffix, nil)
+	now := timeNow.Truncate(time.Millisecond).Unix()
 
 	// Make a large number of children entities.
 	const n = 800
@@ -551,6 +648,7 @@ func TestLargeQuery(t *testing.T) {
 		go func(count, limit, offset, want int) {
 			defer wg.Done()
 
+			ctx := context.WithValue(ctx, ckey{}, fmt.Sprintf("c=%d,l=%d,o=%d", count, limit, offset))
 			// Run iterator through count calls to Next.
 			it := client.Run(ctx, q.Limit(limit).Offset(offset).KeysOnly())
 			for i := 0; i < count; i++ {
@@ -587,30 +685,26 @@ func TestLargeQuery(t *testing.T) {
 			}
 		}(tt.count, tt.limit, tt.offset, tt.want)
 	}
-
 	wg.Wait()
 }
 
-func TestEventualConsistency(t *testing.T) {
+func TestIntegration_EventualConsistency(t *testing.T) {
 	// TODO(jba): either make this actually test eventual consistency, or
 	// delete it. Currently it behaves the same with or without the
 	// EventualConsistency call.
-	if testing.Short() {
-		t.Skip("Integration tests skipped in short mode")
-	}
 	ctx := context.Background()
-	client := newClient(ctx, t)
+	client := newTestClient(ctx, t)
 	defer client.Close()
 
-	parent := NameKey("SQParent", "TestEventualConsistency"+suffix, nil)
-	now := time.Now().Truncate(time.Millisecond).Unix()
+	parent := NameKey("SQParent", "TestIntegration_EventualConsistency"+suffix, nil)
+	now := timeNow.Truncate(time.Millisecond).Unix()
 	children := []*SQChild{
 		{I: 0, T: now, U: now},
 		{I: 1, T: now, U: now},
 		{I: 2, T: now, U: now},
 	}
 	query := NewQuery("SQChild").Ancestor(parent).Filter("T =", now).EventualConsistency()
-	testSmallQueries(t, ctx, client, parent, children, nil, func() {
+	testSmallQueries(ctx, t, client, parent, children, nil, func() {
 		got, err := client.Count(ctx, query)
 		if err != nil {
 			t.Fatalf("Count: %v", err)
@@ -621,16 +715,13 @@ func TestEventualConsistency(t *testing.T) {
 	})
 }
 
-func TestProjection(t *testing.T) {
-	if testing.Short() {
-		t.Skip("Integration tests skipped in short mode")
-	}
+func TestIntegration_Projection(t *testing.T) {
 	ctx := context.Background()
-	client := newClient(ctx, t)
+	client := newTestClient(ctx, t)
 	defer client.Close()
 
-	parent := NameKey("SQParent", "TestProjection"+suffix, nil)
-	now := time.Now().Truncate(time.Millisecond).Unix()
+	parent := NameKey("SQParent", "TestIntegration_Projection"+suffix, nil)
+	now := timeNow.Truncate(time.Millisecond).Unix()
 	children := []*SQChild{
 		{I: 1 << 0, J: 100, T: now, U: now},
 		{I: 1 << 1, J: 100, T: now, U: now},
@@ -639,7 +730,7 @@ func TestProjection(t *testing.T) {
 		{I: 1 << 4, J: 300, T: now, U: now},
 	}
 	baseQuery := NewQuery("SQChild").Ancestor(parent).Filter("T=", now).Filter("J>", 150)
-	testSmallQueries(t, ctx, client, parent, children, []SQTestCase{
+	testSmallQueries(ctx, t, client, parent, children, []SQTestCase{
 		{
 			"project",
 			baseQuery.Project("J"),
@@ -667,12 +758,9 @@ func TestProjection(t *testing.T) {
 	})
 }
 
-func TestAllocateIDs(t *testing.T) {
-	if testing.Short() {
-		t.Skip("Integration tests skipped in short mode")
-	}
+func TestIntegration_AllocateIDs(t *testing.T) {
 	ctx := context.Background()
-	client := newClient(ctx, t)
+	client := newTestClient(ctx, t)
 	defer client.Close()
 
 	keys := make([]*Key, 5)
@@ -693,12 +781,9 @@ func TestAllocateIDs(t *testing.T) {
 	}
 }
 
-func TestGetAllWithFieldMismatch(t *testing.T) {
-	if testing.Short() {
-		t.Skip("Integration tests skipped in short mode")
-	}
+func TestIntegration_GetAllWithFieldMismatch(t *testing.T) {
 	ctx := context.Background()
-	client := newClient(ctx, t)
+	client := newTestClient(ctx, t)
 	defer client.Close()
 
 	type Fat struct {
@@ -712,7 +797,7 @@ func TestGetAllWithFieldMismatch(t *testing.T) {
 	// by default, which prevents a test from being flaky.
 	// See https://cloud.google.com/appengine/docs/go/datastore/queries#Go_Data_consistency
 	// for more information.
-	parent := NameKey("SQParent", "TestGetAllWithFieldMismatch"+suffix, nil)
+	parent := NameKey("SQParent", "TestIntegration_GetAllWithFieldMismatch"+suffix, nil)
 	putKeys := make([]*Key, 3)
 	for i := range putKeys {
 		putKeys[i] = IDKey("GetAllThing", int64(10+i), parent)
@@ -729,10 +814,10 @@ func TestGetAllWithFieldMismatch(t *testing.T) {
 		{X: 22},
 	}
 	getKeys, err := client.GetAll(ctx, NewQuery("GetAllThing").Ancestor(parent), &got)
-	if len(getKeys) != 3 && !reflect.DeepEqual(getKeys, putKeys) {
+	if len(getKeys) != 3 && !testutil.Equal(getKeys, putKeys) {
 		t.Errorf("client.GetAll: keys differ\ngetKeys=%v\nputKeys=%v", getKeys, putKeys)
 	}
-	if !reflect.DeepEqual(got, want) {
+	if !testutil.Equal(got, want) {
 		t.Errorf("client.GetAll: entities differ\ngot =%v\nwant=%v", got, want)
 	}
 	if _, ok := err.(*ErrFieldMismatch); !ok {
@@ -740,12 +825,9 @@ func TestGetAllWithFieldMismatch(t *testing.T) {
 	}
 }
 
-func TestKindlessQueries(t *testing.T) {
-	if testing.Short() {
-		t.Skip("Integration tests skipped in short mode")
-	}
+func TestIntegration_KindlessQueries(t *testing.T) {
 	ctx := context.Background()
-	client := newClient(ctx, t)
+	client := newTestClient(ctx, t)
 	defer client.Close()
 
 	type Dee struct {
@@ -822,54 +904,48 @@ func TestKindlessQueries(t *testing.T) {
 			wantErr: "kind is required for all orders except __key__ ascending",
 		},
 	}
-loop:
-	for _, tc := range testCases {
-		q := tc.query.Ancestor(parent)
-		gotCount, err := client.Count(ctx, q)
-		if err != nil {
-			if tc.wantErr == "" || !strings.Contains(err.Error(), tc.wantErr) {
-				t.Errorf("count %q: err %v, want err %q", tc.desc, err, tc.wantErr)
-			}
-			continue
-		}
-		if tc.wantErr != "" {
-			t.Errorf("count %q: want err %q", tc.desc, tc.wantErr)
-			continue
-		}
-		if gotCount != len(tc.want) {
-			t.Errorf("count %q: got %d want %d", tc.desc, gotCount, len(tc.want))
-			continue
-		}
-		var got []int
-		for iter := client.Run(ctx, q); ; {
-			var dst struct {
-				I          int
-				Why, Pling string
-			}
-			_, err := iter.Next(&dst)
-			if err == iterator.Done {
-				break
-			}
+	for _, test := range testCases {
+		t.Run(test.desc, func(t *testing.T) {
+			q := test.query.Ancestor(parent)
+			gotCount, err := client.Count(ctx, q)
 			if err != nil {
-				t.Errorf("iter.Next %q: %v", tc.desc, err)
-				continue loop
+				if test.wantErr == "" || !strings.Contains(err.Error(), test.wantErr) {
+					t.Fatalf("count %q: err %v, want err %q", test.desc, err, test.wantErr)
+				}
+				return
 			}
-			got = append(got, dst.I)
-		}
-		sort.Ints(got)
-		if !reflect.DeepEqual(got, tc.want) {
-			t.Errorf("elems %q: got %+v want %+v", tc.desc, got, tc.want)
-			continue
-		}
+			if test.wantErr != "" {
+				t.Fatalf("count %q: want err %q", test.desc, test.wantErr)
+			}
+			if gotCount != len(test.want) {
+				t.Fatalf("count %q: got %d want %d", test.desc, gotCount, len(test.want))
+			}
+			var got []int
+			for iter := client.Run(ctx, q); ; {
+				var dst struct {
+					I          int
+					Why, Pling string
+				}
+				_, err := iter.Next(&dst)
+				if err == iterator.Done {
+					break
+				}
+				if err != nil {
+					t.Fatalf("iter.Next %q: %v", test.desc, err)
+				}
+				got = append(got, dst.I)
+			}
+			sort.Ints(got)
+			if !testutil.Equal(got, test.want) {
+				t.Fatalf("elems %q: got %+v want %+v", test.desc, got, test.want)
+			}
+		})
 	}
 }
 
-func TestTransaction(t *testing.T) {
-	if testing.Short() {
-		t.Skip("Integration tests skipped in short mode")
-	}
+func TestIntegration_Transaction(t *testing.T) {
 	ctx := context.Background()
-	client := newClient(ctx, t)
+	client := newTestClient(ctx, t)
 	defer client.Close()
 
 	type Counter struct {
@@ -910,72 +986,113 @@ func TestTransaction(t *testing.T) {
 			wantErr:       ErrConcurrentTransaction,
 		},
 	}
-
-	for i, tt := range tests {
-		// Put a new counter.
-		c := &Counter{N: 10, T: time.Now()}
-		key, err := client.Put(ctx, IncompleteKey("TransCounter", nil), c)
-		if err != nil {
-			t.Errorf("%s: client.Put: %v", tt.desc, err)
-			continue
-		}
-		defer client.Delete(ctx, key)
-
-		// Increment the counter in a transaction.
-		// The test case can manually cause a conflict or return an
-		// error at each attempt.
-		var attempts int
-		_, err = client.RunInTransaction(ctx, func(tx *Transaction) error {
-			attempts++
-			if attempts > len(tt.causeConflict) {
-				return fmt.Errorf("too many attempts. Got %d, max %d", attempts, len(tt.causeConflict))
+	for i, test := range tests {
+		t.Run(test.desc, func(t *testing.T) {
+			// Put a new counter.
+			c := &Counter{N: 10, T: timeNow}
+			key, err := client.Put(ctx, IncompleteKey("TransCounter", nil), c)
+			if err != nil {
+				t.Fatal(err)
 			}
+			defer client.Delete(ctx, key)
 
-			var c Counter
-			if err := tx.Get(key, &c); err != nil {
-				return err
-			}
-			c.N++
-			if _, err := tx.Put(key, &c); err != nil {
-				return err
-			}
+			// Increment the counter in a transaction.
+			// The test case can manually cause a conflict or return an
+			// error at each attempt.
+			var attempts int
+			_, err = client.RunInTransaction(ctx, func(tx *Transaction) error {
+				attempts++
+				if attempts > len(test.causeConflict) {
+					return fmt.Errorf("too many attempts. Got %d, max %d", attempts, len(test.causeConflict))
+				}
 
-			if tt.causeConflict[attempts-1] {
-				c.N += 1
-				if _, err := client.Put(ctx, key, &c); err != nil {
+				var c Counter
+				if err := tx.Get(key, &c); err != nil {
 					return err
 				}
+				c.N++
+				if _, err := tx.Put(key, &c); err != nil {
+					return err
+				}
+
+				if test.causeConflict[attempts-1] {
+					c.N++
+					if _, err := client.Put(ctx, key, &c); err != nil {
+						return err
+					}
+				}
+
+				return test.retErr[attempts-1]
+			}, MaxAttempts(i))
+
+			// Check the error returned by RunInTransaction.
+			if err != test.wantErr {
+				t.Fatalf("got err %v, want %v", err, test.wantErr)
+			}
+			if test.wantErr != nil {
+				// If we were expecting an error, this is where the test ends.
+				return
 			}
 
-			return tt.retErr[attempts-1]
-		}, MaxAttempts(i))
-
-		// Check the error returned by RunInTransaction.
-		if err != tt.wantErr {
-			t.Errorf("%s: got err %v, want %v", tt.desc, err, tt.wantErr)
-			continue
-		}
-		if err != nil {
-			continue
-		}
-
-		// Check the final value of the counter.
-		if err := client.Get(ctx, key, c); err != nil {
-			t.Errorf("%s: client.Get: %v", tt.desc, err)
-			continue
-		}
-		if c.N != tt.want {
-			t.Errorf("%s: counter N=%d, want N=%d", tt.desc, c.N, tt.want)
-		}
+			// Check the final value of the counter.
+			if err := client.Get(ctx, key, c); err != nil {
+				t.Fatal(err)
+			}
+			if c.N != test.want {
+				t.Fatalf("counter N=%d, want N=%d", c.N, test.want)
+			}
+		})
 	}
 }
 
-func TestNilPointers(t *testing.T) {
+func TestIntegration_ReadOnlyTransaction(t *testing.T) {
 	if testing.Short() {
 		t.Skip("Integration tests skipped in short mode")
 	}
 	ctx := context.Background()
-	client := newClient(ctx, t)
+	client := newClient(ctx, t, nil)
+	defer client.Close()
+
+	type value struct{ N int }
+
+	// Put a value.
+	const n = 5
+	v := &value{N: n}
+	key, err := client.Put(ctx, IncompleteKey("roTxn", nil), v)
+	if err != nil {
+		t.Fatal(err)
+	}
+	defer client.Delete(ctx, key)
+
+	// Read it from a read-only transaction.
+	_, err = client.RunInTransaction(ctx, func(tx *Transaction) error {
+		if err := tx.Get(key, v); err != nil {
+			return err
+		}
+		return nil
+	}, ReadOnly)
+	if err != nil {
+		t.Fatal(err)
+	}
+	if v.N != n {
+		t.Fatalf("got %d, want %d", v.N, n)
+	}
+
+	// Attempting to write from a read-only transaction is an error.
+	_, err = client.RunInTransaction(ctx, func(tx *Transaction) error {
+		if _, err := tx.Put(key, v); err != nil {
+			return err
+		}
+		return nil
+	}, ReadOnly)
+	if err == nil {
+		t.Fatal("got nil, want error")
+	}
+}
+
+func TestIntegration_NilPointers(t *testing.T) {
+	ctx := context.Background()
+	client := newTestClient(ctx, t)
 	defer client.Close()
 
 	type X struct {
@@ -993,7 +1110,7 @@ func TestNilPointers(t *testing.T) {
 	xs := make([]*X, 2)
 	if err := client.GetMulti(ctx, keys, xs); err != nil {
 		t.Errorf("GetMulti: %v", err)
-	} else if !reflect.DeepEqual(xs, src) {
+	} else if !testutil.Equal(xs, src) {
 		t.Errorf("GetMulti fetched %v, want %v", xs, src)
 	}
 
@@ -1003,17 +1120,16 @@ func TestNilPointers(t *testing.T) {
 		t.Errorf("Get: err %v; want %v", err, want)
 	}
 
+	// Test that deleting with duplicate keys work.
+	keys = append(keys, keys...)
 	if err := client.DeleteMulti(ctx, keys); err != nil {
 		t.Errorf("Delete: %v", err)
 	}
 }
 
-func TestNestedRepeatedElementNoIndex(t *testing.T) {
-	if testing.Short() {
-		t.Skip("Integration tests skipped in short mode")
-	}
+func TestIntegration_NestedRepeatedElementNoIndex(t *testing.T) {
 	ctx := context.Background()
-	client := newClient(ctx, t)
+	client := newTestClient(ctx, t)
 	defer client.Close()
 
 	type Inner struct {
@@ -1038,3 +1154,181 @@ func TestNestedRepeatedElementNoIndex(t *testing.T) {
 		t.Fatalf("client.Delete: %v", err)
 	}
 }
+
+func TestIntegration_PointerFields(t *testing.T) {
+	ctx := context.Background()
+	client := newTestClient(ctx, t)
+	defer client.Close()
+
+	want := populatedPointers()
+	key, err := client.Put(ctx, IncompleteKey("pointers", nil), want)
+	if err != nil {
+		t.Fatal(err)
+	}
+	var got Pointers
+	if err := client.Get(ctx, key, &got); err != nil {
+		t.Fatal(err)
+	}
+	if got.Pi == nil || *got.Pi != *want.Pi {
+		t.Errorf("Pi: got %v, want %v", got.Pi, *want.Pi)
+	}
+	if got.Ps == nil || *got.Ps != *want.Ps {
+		t.Errorf("Ps: got %v, want %v", got.Ps, *want.Ps)
+	}
+	if got.Pb == nil || *got.Pb != *want.Pb {
+		t.Errorf("Pb: got %v, want %v", got.Pb, *want.Pb)
+	}
+	if got.Pf == nil || *got.Pf != *want.Pf {
+		t.Errorf("Pf: got %v, want %v", got.Pf, *want.Pf)
+	}
+	if got.Pg == nil || *got.Pg != *want.Pg {
+		t.Errorf("Pg: got %v, want %v", got.Pg, *want.Pg)
+	}
+	if got.Pt == nil || !got.Pt.Equal(*want.Pt) {
+		t.Errorf("Pt: got %v, want %v", got.Pt, *want.Pt)
+	}
+}
+
+func TestIntegration_Mutate(t *testing.T) {
+	// test Client.Mutate
+	testMutate(t, func(ctx context.Context, client *Client, muts ...*Mutation) ([]*Key, error) {
+		return client.Mutate(ctx, muts...)
+	})
+	// test Transaction.Mutate
+	testMutate(t, func(ctx context.Context, client *Client, muts ...*Mutation) ([]*Key, error) {
+		var pkeys []*PendingKey
+		commit, err := client.RunInTransaction(ctx, func(tx *Transaction) error {
+			var err error
+			pkeys, err = tx.Mutate(muts...)
+			return err
+		})
+		if err != nil {
+			return nil, err
+		}
+		var keys []*Key
+		for _, pk := range pkeys {
+			keys = append(keys, commit.Key(pk))
+		}
+		return keys, nil
+	})
+}
+
+func testMutate(t *testing.T, mutate func(ctx context.Context, client *Client, muts ...*Mutation) ([]*Key, error)) {
+	ctx := context.Background()
+	client := newTestClient(ctx, t)
+	defer client.Close()
+
+	type T struct{ I int }
+
+	check := func(k *Key, want interface{}) {
+		var x T
+		err := client.Get(ctx, k, &x)
+		switch want := want.(type) {
+		case error:
+			if err != want {
+				t.Errorf("key %s: got error %v, want %v", k, err, want)
+			}
+		case int:
+			if err != nil {
+				t.Fatalf("key %s: %v", k, err)
+			}
+			if x.I != want {
+				t.Errorf("key %s: got %d, want %d", k, x.I, want)
+			}
+		default:
+			panic("check: bad arg")
+		}
+	}
+
+	keys, err := mutate(ctx, client,
+		NewInsert(IncompleteKey("t", nil), &T{1}),
+		NewUpsert(IncompleteKey("t", nil), &T{2}),
+	)
+	if err != nil {
+		t.Fatal(err)
+	}
+	check(keys[0], 1)
+	check(keys[1], 2)
+
+	_, err = mutate(ctx, client,
+		NewUpdate(keys[0], &T{3}),
+		NewDelete(keys[1]),
+	)
+	if err != nil {
+		t.Fatal(err)
+	}
+	check(keys[0], 3)
+	check(keys[1], ErrNoSuchEntity)
+
+	_, err = mutate(ctx, client, NewInsert(keys[0], &T{4}))
+	if got, want := status.Code(err), codes.AlreadyExists; got != want {
+		t.Errorf("Insert existing key: got %s, want %s", got, want)
+	}
+
+	_, err = mutate(ctx, client, NewUpdate(keys[1], &T{4}))
+	if got, want := status.Code(err), codes.NotFound; got != want {
+		t.Errorf("Update non-existing key: got %s, want %s", got, want)
+	}
+}
+
+func TestIntegration_DetectProjectID(t *testing.T) {
+	if testing.Short() {
+		t.Skip("Integration tests skipped in short mode")
+	}
+	ctx := context.Background()
+
+	creds := testutil.Credentials(ctx, ScopeDatastore)
+	if creds == nil {
+		t.Skip("Integration tests skipped. See CONTRIBUTING.md for details")
+	}
+
+	// Use creds with project ID.
+	if _, err := NewClient(ctx, DetectProjectID, option.WithCredentials(creds)); err != nil {
+		t.Errorf("NewClient: %v", err)
+	}
+
+	ts := testutil.ErroringTokenSource{}
+	// Try to use creds without project ID.
+	_, err := NewClient(ctx, DetectProjectID, option.WithTokenSource(ts))
+	if err == nil || err.Error() != "datastore: see the docs on DetectProjectID" {
+		t.Errorf("expected an error while using TokenSource that does not have a project ID")
+	}
+}
+
+var genKeyName = uid.NewSpace("datastore-integration", nil)
+
+func TestIntegration_Project_TimestampStoreAndRetrieve(t *testing.T) {
+	ctx := context.Background()
+	client := newTestClient(ctx, t)
+	defer client.Close()
+
+	type T struct{ Created time.Time }
+
+	keyName := genKeyName.New()
+
+	now := time.Now()
+	k, err := client.Put(ctx, IncompleteKey(keyName, nil), &T{Created: now})
+	if err != nil {
+		t.Fatal(err)
+	}
+	defer func() {
+		if err := client.Delete(ctx, k); err != nil {
+			log.Println(err)
+		}
+	}()
+
+	// Without .Ancestor, this query is eventually consistent (so this test
+	// would be flakey). Ancestor queries, however, are strongly consistent.
+	// See more at https://cloud.google.com/datastore/docs/articles/balancing-strong-and-eventual-consistency-with-google-cloud-datastore/.
+	q := NewQuery(k.Kind).Ancestor(k)
+	res := []T{}
+	if _, err := client.GetAll(ctx, q, &res); err != nil {
+		t.Fatal(err)
+	}
+	if len(res) != 1 {
+		t.Fatalf("expected 1 result, got %d", len(res))
+	}
+	if got, want := res[0].Created.Unix(), now.Unix(); got != want {
+		t.Fatalf("got %v, want %v", got, want)
+	}
+}
