@@ -1,4 +1,4 @@
-// Copyright 4 Google Inc. All Rights Reserved.
+// Copyright 2014 Google LLC
 //
 // Licensed under the Apache License, Version 2.0 (the "License");
 // you may not use this file except in compliance with the License.
@@ -47,27 +47,9 @@ func saveEntity(key *Key, src interface{}) (*pb.Entity, error) {
 	return propertiesToProto(key, props)
 }
 
-// TODO(djd): Convert this and below to return ([]Property, error).
-func saveStructProperty(props *[]Property, name string, opts saveOpts, v reflect.Value) error {
-	p := Property{
-		Name:    name,
-		NoIndex: opts.noIndex,
-	}
-
-	if opts.omitEmpty && isEmptyValue(v) {
-		return nil
-	}
-
-	// First check if field type implements PLS. If so, use PLS to
-	// save.
-	ok, err := plsFieldSave(props, p, name, opts, v)
-	if err != nil {
-		return err
-	}
-	if ok {
-		return nil
-	}
-
+// reflectFieldSave extracts the underlying value of v by reflection,
+// and tries to extract a Property that'll be appended to props.
+func reflectFieldSave(props *[]Property, p Property, name string, opts saveOpts, v reflect.Value) error {
 	switch x := v.Interface().(type) {
 	case *Key, time.Time, GeoPoint:
 		p.Value = x
@@ -81,6 +63,13 @@ func saveStructProperty(props *[]Property, name string, opts saveOpts, v reflect
 			p.Value = v.String()
 		case reflect.Float32, reflect.Float64:
 			p.Value = v.Float()
+
+		case reflect.Interface:
+			// Extract the interface's underlying value and then retry the save.
+			// See issue https://github.com/googleapis/google-cloud-go/issues/1474.
+			v = v.Elem()
+			return reflectFieldSave(props, p, name, opts, v)
+
 		case reflect.Slice:
 			if v.Type().Elem().Kind() == reflect.Uint8 {
 				p.Value = v.Bytes()
@@ -88,9 +77,23 @@ func saveStructProperty(props *[]Property, name string, opts saveOpts, v reflect
 				return saveSliceProperty(props, name, opts, v)
 			}
 		case reflect.Ptr:
+			if isValidPointerType(v.Type().Elem()) {
+				if v.IsNil() {
+					// Nil pointer becomes a nil property value (unless omitempty, handled above).
+					p.Value = nil
+					*props = append(*props, p)
+					return nil
+				}
+				// When we recurse on the derefenced pointer, omitempty no longer applies:
+				// we already know the pointer is not empty, it doesn't matter if its referent
+				// is empty or not.
+				opts.omitEmpty = false
+				return saveStructProperty(props, name, opts, v.Elem())
+			}
 			if v.Type().Elem().Kind() != reflect.Struct {
 				return fmt.Errorf("datastore: unsupported struct field type: %s", v.Type())
 			}
+			// Pointer to struct is a special case.
 			if v.IsNil() {
 				return nil
 			}
@@ -127,6 +130,7 @@ func saveStructProperty(props *[]Property, name string, opts saveOpts, v reflect
 			}
 		}
 	}
+
 	if p.Value == nil {
 		return fmt.Errorf("datastore: unsupported struct field type: %v", v.Type())
 	}
@@ -134,6 +138,30 @@ func saveStructProperty(props *[]Property, name string, opts saveOpts, v reflect
 	return nil
 }
 
+// TODO(djd): Convert this and below to return ([]Property, error).
+func saveStructProperty(props *[]Property, name string, opts saveOpts, v reflect.Value) error {
+	p := Property{
+		Name:    name,
+		NoIndex: opts.noIndex,
+	}
+
+	if opts.omitEmpty && isEmptyValue(v) {
+		return nil
+	}
+
+	// First check if field type implements PLS. If so, use PLS to
+	// save.
+	ok, err := plsFieldSave(props, p, name, opts, v)
+	if err != nil {
+		return err
+	}
+	if ok {
+		return nil
+	}
+
+	return reflectFieldSave(props, p, name, opts, v)
+}
+
 // plsFieldSave first tries to converts v's value to a PLS, then v's addressed
 // value to a PLS. If neither succeeds, plsFieldSave returns false for first return
 // value.
@@ -296,7 +324,7 @@ func propertiesToProto(key *Key, props []Property) (*pb.Entity, error) {
 	}
 	indexedProps := 0
 	for _, p := range props {
-		// Do not send a Key value a a field to datastore.
+		// Do not send a Key value a field to datastore.
 		if p.Name == keyFieldName {
 			continue
 		}
@@ -395,10 +423,18 @@ func interfaceToProto(iv interface{}, noIndex bool) (*pb.Value, error) {
 		// than the top-level value.
 		val.ExcludeFromIndexes = false
 	default:
-		if iv != nil {
-			return nil, fmt.Errorf("invalid Value type %t", iv)
+		rv := reflect.ValueOf(iv)
+		if !rv.IsValid() {
+			val.ValueType = &pb.Value_NullValue{}
+		} else if rv.Kind() == reflect.Ptr { // non-nil pointer: dereference
+			if rv.IsNil() {
+				val.ValueType = &pb.Value_NullValue{}
+				return val, nil
+			}
+			return interfaceToProto(rv.Elem().Interface(), noIndex)
+		} else {
+			return nil, fmt.Errorf("invalid Value type %T", iv)
 		}
-		val.ValueType = &pb.Value_NullValue{}
 	}
 	// TODO(jbd): Support EntityValue.
 	return val, nil
@@ -420,6 +456,29 @@ func isEmptyValue(v reflect.Value) bool {
 		return v.Float() == 0
 	case reflect.Interface, reflect.Ptr:
 		return v.IsNil()
+	case reflect.Struct:
+		if t, ok := v.Interface().(time.Time); ok {
+			return t.IsZero()
+		}
+	}
+	return false
+}
+
+// isValidPointerType reports whether a struct field can be a pointer to type t
+// for the purposes of saving and loading.
+func isValidPointerType(t reflect.Type) bool {
+	if t == typeOfTime || t == typeOfGeoPoint {
+		return true
+	}
+	switch t.Kind() {
+	case reflect.Int, reflect.Int8, reflect.Int16, reflect.Int32, reflect.Int64:
+		return true
+	case reflect.Bool:
+		return true
+	case reflect.String:
+		return true
+	case reflect.Float32, reflect.Float64:
+		return true
 	}
 	return false
 }
