@@ -1,4 +1,4 @@
-// Copyright 2016 Google Inc. All Rights Reserved.
+// Copyright 2016 Google LLC
 //
 // Licensed under the Apache License, Version 2.0 (the "License");
 // you may not use this file except in compliance with the License.
@@ -26,9 +26,8 @@
 package logadmin // import "cloud.google.com/go/logging/logadmin"
 
 import (
-	"errors"
+	"context"
 	"fmt"
-	"math"
 	"net/http"
 	"net/url"
 	"strings"
@@ -39,35 +38,32 @@ import (
 	vkit "cloud.google.com/go/logging/apiv2"
 	"cloud.google.com/go/logging/internal"
 	"github.com/golang/protobuf/ptypes"
-	gax "github.com/googleapis/gax-go"
-	"golang.org/x/net/context"
+	gax "github.com/googleapis/gax-go/v2"
 	"google.golang.org/api/iterator"
 	"google.golang.org/api/option"
+	_ "google.golang.org/genproto/googleapis/appengine/logging/v1" // Import the following so EntryIterator can unmarshal log protos.
+	_ "google.golang.org/genproto/googleapis/cloud/audit"
 	logtypepb "google.golang.org/genproto/googleapis/logging/type"
 	logpb "google.golang.org/genproto/googleapis/logging/v2"
 	"google.golang.org/grpc/codes"
-	// Import the following so EntryIterator can unmarshal log protos.
-	_ "google.golang.org/genproto/googleapis/cloud/audit"
 )
 
 // Client is a Logging client. A Client is associated with a single Cloud project.
 type Client struct {
-	lClient   *vkit.Client        // logging client
-	sClient   *vkit.ConfigClient  // sink client
-	mClient   *vkit.MetricsClient // metric client
-	projectID string
-	closed    bool
+	lClient *vkit.Client        // logging client
+	sClient *vkit.ConfigClient  // sink client
+	mClient *vkit.MetricsClient // metric client
+	parent  string
+	closed  bool
 }
 
 // NewClient returns a new logging client associated with the provided project ID.
 //
 // By default NewClient uses AdminScope. To use a different scope, call
 // NewClient using a WithScopes option (see https://godoc.org/google.golang.org/api/option#WithScopes).
-func NewClient(ctx context.Context, projectID string, opts ...option.ClientOption) (*Client, error) {
-	// Check for '/' in project ID to reserve the ability to support various owning resources,
-	// in the form "{Collection}/{Name}", for instance "organizations/my-org".
-	if strings.ContainsRune(projectID, '/') {
-		return nil, errors.New("logging: project ID contains '/'")
+func NewClient(ctx context.Context, parent string, opts ...option.ClientOption) (*Client, error) {
+	if !strings.ContainsRune(parent, '/') {
+		parent = "projects/" + parent
 	}
 	opts = append([]option.ClientOption{
 		option.WithEndpoint(internal.ProdAddr),
@@ -104,19 +100,14 @@ func NewClient(ctx context.Context, projectID string, opts ...option.ClientOptio
 	sc.SetGoogleClientInfo("gccl", version.Repo)
 	mc.SetGoogleClientInfo("gccl", version.Repo)
 	client := &Client{
-		lClient:   lc,
-		sClient:   sc,
-		mClient:   mc,
-		projectID: projectID,
+		lClient: lc,
+		sClient: sc,
+		mClient: mc,
+		parent:  parent,
 	}
 	return client, nil
 }
 
-// parent returns the string used in many RPCs to denote the parent resource of the log.
-func (c *Client) parent() string {
-	return "projects/" + c.projectID
-}
-
 // Close closes the client.
 func (c *Client) Close() error {
 	if c.closed {
@@ -135,7 +126,7 @@ func (c *Client) Close() error {
 // logID identifies the log within the project. An example log ID is "syslog". Requires AdminScope.
 func (c *Client) DeleteLog(ctx context.Context, logID string) error {
 	return c.lClient.DeleteLog(ctx, &logpb.DeleteLogRequest{
-		LogName: internal.LogPath(c.parent(), logID),
+		LogName: internal.LogPath(c.parent, logID),
 	})
 }
 
@@ -195,6 +186,16 @@ func (p projectIDs) set(r *logpb.ListLogEntriesRequest) {
 	}
 }
 
+// ResourceNames sets the resource names from which to retrieve
+// log entries. Examples: "projects/my-project-1A", "organizations/my-org".
+func ResourceNames(rns []string) EntriesOption { return resourceNames(rns) }
+
+type resourceNames []string
+
+func (rn resourceNames) set(r *logpb.ListLogEntriesRequest) {
+	r.ResourceNames = append([]string(nil), rn...)
+}
+
 // Filter sets an advanced logs filter for listing log entries (see
 // https://cloud.google.com/logging/docs/view/advanced_filters). The filter is
 // compared against all log entries in the projects specified by ProjectIDs.
@@ -226,7 +227,7 @@ func (newestFirst) set(r *logpb.ListLogEntriesRequest) { r.OrderBy = "timestamp
 // NewClient. This may be overridden by passing a ProjectIDs option. Requires ReadScope or AdminScope.
 func (c *Client) Entries(ctx context.Context, opts ...EntriesOption) *EntryIterator {
 	it := &EntryIterator{
-		it: c.lClient.ListLogEntries(ctx, listLogEntriesRequest(c.projectID, opts)),
+		it: c.lClient.ListLogEntries(ctx, listLogEntriesRequest(c.parent, opts)),
 	}
 	it.pageInfo, it.nextFunc = iterator.NewPageInfo(
 		it.fetch,
@@ -235,9 +236,9 @@ func (c *Client) Entries(ctx context.Context, opts ...EntriesOption) *EntryItera
 	return it
 }
 
-func listLogEntriesRequest(projectID string, opts []EntriesOption) *logpb.ListLogEntriesRequest {
+func listLogEntriesRequest(parent string, opts []EntriesOption) *logpb.ListLogEntriesRequest {
 	req := &logpb.ListLogEntriesRequest{
-		ResourceNames: []string{"projects/" + projectID},
+		ResourceNames: []string{parent},
 	}
 	for _, opt := range opts {
 		opt.set(req)
@@ -283,13 +284,6 @@ func (it *EntryIterator) fetch(pageSize int, pageToken string) (string, error) {
 	})
 }
 
-func trunc32(i int) int32 {
-	if i > math.MaxInt32 {
-		i = math.MaxInt32
-	}
-	return int32(i)
-}
-
 var slashUnescaper = strings.NewReplacer("%2F", "/", "%2f", "/")
 
 func fromLogEntry(le *logpb.LogEntry) (*logging.Entry, error) {
@@ -314,6 +308,9 @@ func fromLogEntry(le *logpb.LogEntry) (*logging.Entry, error) {
 		// TODO(jba): convert to map[string]interface{}?
 		payload = x.JsonPayload
 
+	case nil:
+		payload = nil
+
 	default:
 		return nil, fmt.Errorf("logging: unknown payload type: %T", le.Payload)
 	}
@@ -322,24 +319,25 @@ func fromLogEntry(le *logpb.LogEntry) (*logging.Entry, error) {
 		return nil, err
 	}
 	return &logging.Entry{
-		Timestamp:   time,
-		Severity:    logging.Severity(le.Severity),
-		Payload:     payload,
-		Labels:      le.Labels,
-		InsertID:    le.InsertId,
-		HTTPRequest: hr,
-		Operation:   le.Operation,
-		LogName:     slashUnescaper.Replace(le.LogName),
-		Resource:    le.Resource,
-		Trace:       le.Trace,
+		Timestamp:      time,
+		Severity:       logging.Severity(le.Severity),
+		Payload:        payload,
+		Labels:         le.Labels,
+		InsertID:       le.InsertId,
+		HTTPRequest:    hr,
+		Operation:      le.Operation,
+		LogName:        slashUnescaper.Replace(le.LogName),
+		Resource:       le.Resource,
+		Trace:          le.Trace,
+		SourceLocation: le.SourceLocation,
 	}, nil
 }
 
 // Logs lists the logs owned by the parent resource of the client.
 func (c *Client) Logs(ctx context.Context) *LogIterator {
 	it := &LogIterator{
-		parentResource: c.parent(),
-		it:             c.lClient.ListLogs(ctx, &logpb.ListLogsRequest{Parent: c.parent()}),
+		parentResource: c.parent,
+		it:             c.lClient.ListLogs(ctx, &logpb.ListLogsRequest{Parent: c.parent}),
 	}
 	it.pageInfo, it.nextFunc = iterator.NewPageInfo(
 		it.fetch,
