@@ -1,4 +1,4 @@
-// Copyright 2014 Google Inc. All Rights Reserved.
+// Copyright 2014 Google LLC
 //
 // Licensed under the Apache License, Version 2.0 (the "License");
 // you may not use this file except in compliance with the License.
@@ -15,10 +15,10 @@
 package storage
 
 import (
+	"context"
 	"errors"
 	"testing"
 
-	"golang.org/x/net/context"
 	"google.golang.org/api/googleapi"
 )
 
